/*
 * $OpenBSD$
 * Copyright (c) 2011-2012 Marc Espie <espie@openbsd.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sndio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <stdint.h>
#include <getopt.h>

#define BUF_SIZE 1024

struct cb_data
{
	int delta;
	int increment;
	int scroll;
};

void 
cb(void *p, int delta)
{
	struct cb_data *data = p;

	if (!data->scroll)
		return;
	data->delta += delta;
	while (data->delta >= data->increment) {
		size_t len;
		char *s = fgetln(stdin, &len);
		if (s)
			write(STDOUT_FILENO, s, len);
		else {
			data->scroll = 0;
			break;
	    	}
		data->delta -= data->increment;
	}
}

void __dead 
play(struct sio_hdl *io, struct sio_par *par, int bpm, int frac, int tone, 
    int beats, int tone2)
{
	int16_t sine[BUF_SIZE];
	double spos = 0; /* sine tone counter */
	double sstep = tone * 2 * M_PI / par->rate; /* and step */
	double sstep2 = tone2 * 2 * M_PI / par->rate;
	double bpos = 0; /* bpm counter */
	double bstep = bpm * frac / 60.0 / par->rate; /* and step */
	int beat = beats ? beats : 1;

	while (1) {
		unsigned int i; /* fill buffer */
		for (i = 0; i < BUF_SIZE; i++) {
			bpos += bstep;
			if (bpos > frac) {
				bpos -= frac;
				beat++;
			}
			if (bpos < 1) {
				sine[i] = INT16_MAX * sin(spos);
				if (beat == beats)
					beat = 0;
				if (beat == 0)
					spos += sstep2;
				else
					spos += sstep;
			} else {
				sine[i] = sine[(i-1)%BUF_SIZE]/2;
				spos = 0;
			}
			if (par->pchan == 2) {
				sine[i] = sine[i-1];
				i++;
			}
		}
		/* if we can't write it, what can we do ? */
		(void)sio_write(io, sine, sizeof sine);
	}
}

void
bail(const char *msg)
{
	fprintf(stderr, "%s\n", msg);
	exit(1);
}

void
usage()
{
	fprintf(stderr, 
"Usage: metronome [-F] [-b freq] [-f int] [-s freq] [-t freq] [-B int] [-T freq]\n");
	exit(1);
}

int
main(int argc, char *argv[])
{
	struct sio_hdl *io;
	struct sio_par par;
	struct cb_data d;

	int bpm = 0, frac = 0, tone = 0, beats = 0, tone2 = 0, sampling = 0;

	int ch;

	d.scroll = 0;

	/* parameters, then default values */
	while ((ch = getopt(argc, argv, "b:f:Fs:t:B:T:")) != -1) {
		switch (ch) {
		case 'b':
			bpm = atoi(optarg); 
			break;
		case 'f':
			frac = atoi(optarg); 
			break;
		case 't':
			tone = atoi(optarg);
			break;
		case 'B':
			beats = atoi(optarg);
			break;
		case 'T':
			tone2 = atoi(optarg);
			break;
		case 's':
			sampling = atoi(optarg);
			break;
		case 'F':
			d.scroll = 1;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 0)
		usage();

	if (!bpm)
		bpm = 60;
	if (!frac)
		frac = 10;
	if (!tone)
		tone = 140;
	if (!tone2)
		tone2 = 2 * tone;
	if (!sampling)
		sampling = 48000;

	io = sio_open(NULL, SIO_PLAY, 0);
	if (!io)
		bail("sio_open");
	sio_initpar(&par);
	par.bits = 16;
	par.sig = 1;
	par.pchan = 1;
	par.rate = sampling;
	if (sio_setpar(io, &par) == 0)
		bail("sio_setpar");
	if (sio_getpar(io, &par) == 0)
		bail("sio_getpar");
	if (par.bits != 16)
		bail("Can't do 16 bits");
	if (par.sig != 1)
		bail("Can't do signed 16 bits");
	if (par.pchan > 2)
		bail("Can only do mono or stereo");
	if (d.scroll) {
		d.increment = par.rate * 60.0 / bpm;
		d.delta = d.increment;
		sio_onmove(io, cb, &d);
	}
	if (sio_start(io) == 0)
		bail("sio_start");
	play(io, &par, bpm, frac, tone, beats, tone2);
	/*NOTREACHED*/
}
